// Copyright © 2020 Amazon Web Services
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package org.opengroup.osdu.register.provider.aws.subscriber;

import com.amazonaws.services.dynamodbv2.datamodeling.DynamoDBDeleteExpression;
import com.amazonaws.services.dynamodbv2.datamodeling.PaginatedQueryList;
import com.amazonaws.services.dynamodbv2.model.AttributeValue;
import com.amazonaws.services.dynamodbv2.model.ConditionalCheckFailedException;
import com.amazonaws.services.sns.AmazonSNS;
import com.google.cloud.Timestamp;
import org.opengroup.osdu.core.aws.dynamodb.DynamoDBQueryHelper;
import org.opengroup.osdu.core.aws.sns.AmazonSNSConfig;
import org.opengroup.osdu.core.aws.ssm.ParameterStorePropertySource;
import org.opengroup.osdu.core.aws.ssm.SSMConfig;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.aws.action.ActionDoc;
import org.opengroup.osdu.register.provider.aws.config.AwsServiceConfig;
import org.opengroup.osdu.register.provider.aws.util.DocUtil;
import org.opengroup.osdu.register.provider.interfaces.subscriber.ISubscriptionRepository;
import org.opengroup.osdu.register.subscriber.model.*;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import javax.annotation.PostConstruct;
import javax.inject.Inject;
import java.nio.ByteBuffer;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

@Repository
public class AwsSubscriptionRepo implements ISubscriptionRepository {

    @Autowired
    private DpsHeaders dpsHeaders;

    @Autowired
    private JaxRsDpsLog logger;

    private DynamoDBQueryHelper queryHelper;

    @Inject
    private AwsServiceConfig serviceConfig;


    private String amazonSNSTopicArn_records_changed;

    private  SSMConfig ssmConfig;

    private   ParameterStorePropertySource ssm;

    @Autowired
    private SubscriptionHelper snsHelper;

    @Autowired
    private KmsHelper kmsHelper;


    @Autowired
    DocUtil docUtil;

    @PostConstruct
    public void init() {
        // TODO: serviceConfig.environment isn't correct and needs to be table prefix. Maybe the "-" will fix it
        queryHelper = new DynamoDBQueryHelper(serviceConfig.getDynamoDbEndpoint(),
                serviceConfig.getAmazonRegion(),
                serviceConfig.getDynamoDbTablePrefix());
        amazonSNSTopicArn_records_changed=serviceConfig.getSnsTopicArn();
    }

    @Override
    public Subscription create(Subscription s) throws Exception {


        String pushEndPoint = s.getPushEndpoint();
        String subscriptionArn = snsHelper.createPushSubscription(amazonSNSTopicArn_records_changed,pushEndPoint);
        ByteBuffer encryptedSecretValue = kmsHelper.encrypt(s.getSecret().toString());
        SubscriptionDoc doc = SubscriptionDoc.mapFrom(s, dpsHeaders.getPartitionId(),subscriptionArn,encryptedSecretValue);

        try {
            queryHelper.saveWithHashCondition(doc,docUtil.getHashKey());
        }
        catch(ConditionalCheckFailedException e){
            logger.error(String.format("A subscription already exists with the id: %s", s.getId()));
            throw new AppException(409, "Conflict", String.format("A subscription  already exists with the id: %s", s.getId()));
        }
        catch(AppException e)
        {
            throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
        }
        return s;

    }

    @Override
    public Subscription get(String id) {
        SubscriptionDoc doc = queryHelper.loadByPrimaryKey(SubscriptionDoc.class, id);
        if (doc == null) {
            logger.error(String.format("Subscription with id %s does not exist.", id));
            throw new AppException(404, "Not found", String.format("Subscription with id %s does not exist.", id));
        } else {
            String secretValue = kmsHelper.decrypt(doc.getSecretValue());
            Secret secret = docUtil.createSecret(doc.getSecretType(),secretValue);
            Subscription s =  SubscriptionDoc.mapTo(doc,secret);

            return s;
        }
    }

    @Override
    public List<Subscription> query(String notificationId) {
        String key=String.format("%s:%s",dpsHeaders.getPartitionId(),notificationId);

        SubscriptionDoc gsiQuery = new SubscriptionDoc();
        gsiQuery.setPartitionIdNotificationId(key);

        List<Subscription> subsList = docUtil.getSubscriptionList(queryHelper,gsiQuery);

        return subsList;


    }

    @Override
    public boolean delete(String id) {

        SubscriptionDoc doc=null;
        String snsSubscriptionArn="";

        try {
            //Need this to retrieve the subscription Arn
             doc = queryHelper.loadByPrimaryKey(SubscriptionDoc.class, id);
            if (doc == null) {
                logger.error(String.format("Subscription with id %s does not exist.", id));
                throw new AppException(404, "Not found", String.format("Subscription with id %s does not exist.", id));
            }
            else {
                snsSubscriptionArn = doc.getSnssubscriptionArn();
               // queryHelper.deleteByPrimaryKey(SubscriptionDoc.class, id);
                SubscriptionDoc objectTodelete = new SubscriptionDoc();
                objectTodelete.setId(id);
                DynamoDBDeleteExpression deleteExpression = new DynamoDBDeleteExpression()
                        .withConditionExpression("attribute_exists(id)");
                queryHelper.deleteByObjectWithCondition(objectTodelete,deleteExpression);
            }
        }
        catch(ConditionalCheckFailedException e){
            logger.error(String.format("Failed to delete: %s Object does not exist",id));
            throw new AppException(404, "ObjectNotFound", String.format("The Subscription you are trying to delete: %s does not exist", id));
        }
        catch(AppException e)
        {
            throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
        }


        //delete the SNS subscription
        //this will throw 404 if SNS Subscription not found
        snsHelper.deletePushSubscription(snsSubscriptionArn);
        return true;

    }

    @Override
    public boolean patch(Subscription subscription, Secret secret) {
        SubscriptionDoc doc=null;
        String id = subscription.getId();
        try {
            doc = queryHelper.loadByPrimaryKey(SubscriptionDoc.class, id);
        }catch(AppException e)
        {
            if(e.getError().getCode() == 404) {
                logger.error("Could not find subscription with Id %s ", id);
                throw e;
            }
            else {
                logger.error(e.getMessage());
                throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
            }
        }
        //update the secret
        doc.setSecretType(secret.getSecretType());
        doc.setSecretValue(kmsHelper.encrypt(secret.toString()));
      try{
            queryHelper.save(doc);

        } catch (AppException e) {
          String msg="Error while saving the updated secret";
          throw new AppException(e.getError().getCode(), e.getError().getReason(), msg);
        }
        return true;
    }

    @Override
    public List<Subscription> getAll() throws Exception {

        List<String> filters = new ArrayList<>();
        Map<String, AttributeValue> valueMap = new HashMap<>();
        List<SubscriptionDoc> results=null;

        filters.add("dataPartitionId = :dataPartitionId");
        valueMap.put(":dataPartitionId", new AttributeValue().withS(dpsHeaders.getPartitionId()));

        String filterExpression = String.join(" and ", filters);
        logger.info(String.format("Subscription query filter expression: %s", filterExpression));

        try {

           results = queryHelper.scanTable(SubscriptionDoc.class, filterExpression, valueMap);
        }
        catch(AppException e)
        {
            String msg="Error while getting ALL subscriptions";
            throw new AppException(e.getError().getCode(), e.getError().getReason(), msg);
        }

        List<Subscription> subsList = new ArrayList<Subscription>();
        for (SubscriptionDoc subsDoc : results){
            String secretValue = kmsHelper.decrypt(subsDoc.getSecretValue());
            Secret secret = docUtil.createSecret(subsDoc.getSecretType(),secretValue);
            subsList.add(SubscriptionDoc.mapTo(subsDoc,secret));

        }

        return subsList;
    }

}
