// Copyright © Microsoft Corporation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package org.opengroup.osdu.register.provider.azure.subscriber;

import com.azure.cosmos.FeedOptions;
import com.azure.cosmos.SqlParameter;
import com.azure.cosmos.SqlParameterList;
import com.azure.cosmos.SqlQuerySpec;
import com.google.cloud.Timestamp;
import org.opengroup.osdu.azure.CosmosStore;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.provider.azure.di.AzureBootstrapConfig;
import org.opengroup.osdu.register.provider.azure.di.CosmosContainerConfig;
import org.opengroup.osdu.register.provider.azure.util.CryptographyUtil;
import org.opengroup.osdu.register.provider.interfaces.subscriber.ISubscriptionRepository;
import org.opengroup.osdu.register.subscriber.model.GsaSecret;
import org.opengroup.osdu.register.subscriber.model.GsaSecretValue;
import org.opengroup.osdu.register.subscriber.model.HmacSecret;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.subscriber.model.Subscription;
import org.opengroup.osdu.register.utils.Constants;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

@Repository
public class SubscriptionRepository implements ISubscriptionRepository {

    @Autowired
    private AzureBootstrapConfig azureBootstrapConfig;

    @Autowired
    private CosmosContainerConfig cosmosContainerConfig;

    @Autowired
    private CosmosStore cosmosStore;

    @Autowired
    private DpsHeaders dpsHeaders;

    @Autowired
    private JaxRsDpsLog logger;

    @Autowired
    private CryptographyUtil cryptographyUtil;

    @Autowired
    private PushSubscription pushSubscription;

    /**
     * @param input     Subscription Object
     * @return          Subscription Object
     *
     *  This method creates a Record for the subscription in CosmosDb along
     *  with a Push Subscription in Event Grid.
     *
     *  Expected Flow -> Create a record in Cosmos Db along with a Push Subscription in Event Grid
     *  Possible Failures ->
     *  1) Creating PushSubscription in Event Grid
     *     How it is handled -> Delete the record in Cosmos Db as well
     *     Possible Issue -> What happens if this delete operation on Cosmos Db fails?
     *     Users won't be able to create a subscription again with same Topic and
     *     push endpoint combination since system will generate 409 on next retry
     *     because of presence of record with same Id in Cosmos Db
     *
     *     How to resolve this?
     *     createPushSubscriptionIfDoesNotExist method takes care of that
     *     When Cosmos Db throws 409 while creating a new record we can check if a Push Subscription
     *     for that Record already exists in Event Grid or not.
     *     If present -> Return 409
     *     If not present -> We can try creating the Push Subscription again in Event Grid
     *                       Also we will update the record in Cosmos Db in case other fields might
     *                       be different like description etc.
     *
     */

    @Override
    public Subscription create(Subscription input) {

        SubscriptionDoc doc = new SubscriptionDoc(input, dpsHeaders.getPartitionId());
        String encryptedSecret = cryptographyUtil.encryptData(input.getSecret().toString());
        doc.setSecretValue(encryptedSecret);
        doc.setKeyName(cryptographyUtil.getKeyName());
        doc.setKeyVersion(cryptographyUtil.getKeyVersion());

        try {
            cosmosStore.createItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), doc);
            logger.info(String.format("Record successfully created for Subscription with id %s", input.getId()));
        }
        catch (AppException e) {
            if(e.getError().getCode() == 409) {
                // This flow is added to handle the scenario where creating PushSubscription in EventGrid failed
                // and deleting the corresponding record from Cosmos Db was also unsuccessful.
                // This will result in an 500 Exception so the user should be able to create the subscription with
                // the same topic and pushEndpoint combination again
                logger.info(String.format("Record already exists for Subscription with id %s", input.getId()));
                SubscriptionDoc output = createPushSubscriptionIfDoesNotExist(doc);
                input.setNotificationId(output.getNotificationId());
                return input;
            }
            else {
                logger.error(e.getMessage());
                throw new AppException(e.getError().getCode(), e.getError().getReason(), e.getMessage());
            }
        }

        try {
            pushSubscription.createPushSubscription(doc.getNotificationId(), doc.getTopic());
            logger.info("Push Subscription created with Event Grid ID:" + doc.getNotificationId());
            return input;
        }
        catch(AppException e) {
            if(e.getError().getCode() == 409) {
                throw new AppException(409, "Conflict", "Another request is trying to create the same subscription");
            }
            try {
                cosmosStore.deleteItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), input.getId(), dpsHeaders.getPartitionId());
                logger.info("Record deleted for subscription with ID: " + input.getId());
            }
            catch (AppException ignored) {}

            logger.error("Unexpected error creating subscription");
            throw new AppException(500, "Server Error", "Unexpected error creating subscription");
        }
    }

    @Override
    public Subscription get(String id) {

        SubscriptionDoc doc = cosmosStore.findItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), id, dpsHeaders.getPartitionId(), SubscriptionDoc.class)
                .orElseThrow(() -> {
                    logger.error(String.format("Subscriber with id %s does not exist.", id));
                    return new AppException(404, "Not found", String.format("Subscriber with id %s does not exist.", id));
                });

        return convertToSubscriptionClass(doc);
    }

    @Override
    public List<Subscription> query(String notificationId) {

        SqlQuerySpec query = new SqlQuerySpec("SELECT * from c where c.notificationId = @notificationId and c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@notificationId", notificationId));
        pars.add(new SqlParameter("@partitionId", dpsHeaders.getPartitionId()));

        FeedOptions options = new FeedOptions();

        return cosmosStore.queryItems(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), query, options, SubscriptionDoc.class).stream()
                .map(this::convertToSubscriptionClass)
                .collect(Collectors.toList());
    }

    /**
     * @param id    Identifier for the subscription
     * @return      true if subscription successfully deleted
     *              false if subscription not found
     *              Exception thrown -> AppException with Status Codes 409 and 500
     */
    @Override
    public boolean delete(String id) {

        Optional<SubscriptionDoc> subscription = cosmosStore.findItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), id, dpsHeaders.getPartitionId(), SubscriptionDoc.class);

        if(!subscription.isPresent()){
            logger.info("Record not found for subscription with ID: " + id);
            return false;
        }
        logger.info("Record found for subscription with ID: " + id);
        try {
            pushSubscription.deletePushSubscription(subscription.get().getNotificationId(), subscription.get().getTopic());
            logger.info("Push Subscription deleted with Event Grid ID:" + subscription.get().getNotificationId());
            try {
                cosmosStore.deleteItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), id, dpsHeaders.getPartitionId());
                logger.info("Record deleted for subscription with ID: " + id);
                return true;
            }
            catch (AppException e) {
                if(e.getError().getCode() == 404) {
                    logger.info(String.format("Subscriber with id %s does not exist.", id));
                    return false;
                }
                else {
                    throw e;
                }
            }
        }
        catch (AppException e) {
            if(e.getError().getCode() == 409) {
                logger.error("Another request is trying to delete the same subscription");
                throw new AppException(409, "Conflict", "Another request is trying to delete the same subscription");
            }
            else {
                logger.error("Unexpected error deleting subscription");
                throw new AppException(500, "Server Error", "Unexpected error deleting subscription");
            }
        }

    }

    @Override
    public boolean patch(Subscription subscription, Secret secret){
        return updateSecret(subscription, secret);
    }

    @Override
    public List<Subscription> getAll() {

        SqlQuerySpec query = new SqlQuerySpec("SELECT * from c where c.dataPartitionId = @partitionId");

        SqlParameterList pars = query.getParameters();
        pars.add(new SqlParameter("@partitionId", dpsHeaders.getPartitionId()));

        FeedOptions options = new FeedOptions();

        return cosmosStore.queryItems(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), query, options, SubscriptionDoc.class).stream()
                .map(this::convertToSubscriptionClass)
                .collect(Collectors.toList());
    }

    private SubscriptionDoc createPushSubscriptionIfDoesNotExist(SubscriptionDoc input) {
        try {
            // We are fetching the record from cosmos db again because we want to use the original
            // notification-id to create the Push Subscription
            Optional<SubscriptionDoc> originalDoc = cosmosStore.findItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), input.getId(), dpsHeaders.getPartitionId(), SubscriptionDoc.class);
            originalDoc.ifPresent(subscriptionDoc -> input.setNotificationId(subscriptionDoc.getNotificationId()));
            logger.info(String.format("Creating Push Subscription with id %s if does not exists", input.getNotificationId()));

            // We will check if Push Subscription does not exist then we should try creating it again since
            // the corresponding record in the Cosmos Db is already present
            pushSubscription.checkIfPushSubscriptionExists(input.getNotificationId(), input.getTopic());
            logger.error("A subscriber already exists with the same topic and endpoint combination");
            throw new AppException(409, "Conflict", "A subscriber already exists with the same topic and endpoint combination");
        }
        catch (AppException e) {
            if(e.getError().getCode() == 409) {
                throw e;
            }
            else if(e.getError().getCode() == 404){
                try {
                    // If Push Subscription is not yet created we should create it, also we might
                    // need to update the cosmos db record since some other fields might be different
                    cosmosStore.upsertItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), input);
                    logger.info("Updating Record for subscription with ID: ", input.getId());
                    // Now creating the Push Subscription, if it fails again we should throw server error
                    // otherwise we can return Success code as response
                    pushSubscription.createPushSubscription(input.getNotificationId(), input.getTopic());
                    logger.info("Push Subscription created with Event Grid ID:" + input.getNotificationId());
                    return input;
                }
                catch (AppException exception) {
                    if(exception.getError().getCode() == 409) {
                        logger.error("Another request is trying to create the same Push subscription");
                        throw new AppException(409, "Conflict", "Another request is trying to create the same Push subscription");
                    }
                    else {
                        logger.error("Unexpected error creating subscription");
                        throw new AppException(500, "Server Error", "Unexpected error creating subscription");
                    }
                }
            }
            else {
                logger.error("Unexpected error creating subscription");
                throw new AppException(500, "Server Error", "Unexpected error creating subscription");
            }
        }
    }

    private boolean updateSecret(Subscription subscription, Secret secret) {
        SubscriptionDoc doc = new SubscriptionDoc(subscription, dpsHeaders.getPartitionId());
        doc.setSecretType(secret.getSecretType());
        doc.setSecretValue(cryptographyUtil.encryptData(secret.toString()));

        cosmosStore.upsertItem(dpsHeaders.getPartitionId(), azureBootstrapConfig.getCosmosDBName(), cosmosContainerConfig.getSubscriptionContainerName(), doc);
        return true;
    }

    private Subscription convertToSubscriptionClass(SubscriptionDoc doc) {

        String secretValue = cryptographyUtil.decryptData(doc.getSecretValue());
        Secret secret;
        if (doc.getSecretType().equals(Constants.GSASecret)) {
            GsaSecret gsaSecret = new GsaSecret();
            String[] splitSecret = secretValue.split("`");
            gsaSecret.setValue(new GsaSecretValue(splitSecret[0], splitSecret[1]));
            secret = gsaSecret;
        } else {
            HmacSecret hmacSecret = new HmacSecret();
            hmacSecret.setValue(secretValue);
            secret = hmacSecret;
        }
        return new Subscription(doc.getId(), doc.getName(),doc.getDescription(), doc.getTopic(), doc.getPushEndpoint(),
                doc.getCreatedBy(), Timestamp.of(doc.getCreatedOnEpoch()), doc.getNotificationId(), secret);
    }
}
