/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.api;

import com.fasterxml.jackson.databind.JsonNode;
import org.opengroup.osdu.register.action.model.Action;
import org.opengroup.osdu.register.action.model.CreateActionDto;
import org.opengroup.osdu.register.action.model.ParsedAction;
import org.opengroup.osdu.register.action.model.TestActionRequest;
import org.opengroup.osdu.register.provider.interfaces.action.IActionRepo;
import org.opengroup.osdu.register.provider.interfaces.action.IRetrieveActionService;
import org.opengroup.osdu.register.action.services.TestActionService;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.register.logging.AuditLog;
import org.opengroup.osdu.register.utils.ServiceRole;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.context.annotation.RequestScope;

import javax.validation.Valid;
import javax.validation.constraints.NotBlank;
import javax.validation.constraints.Pattern;
import java.util.List;

import static org.apache.http.HttpStatus.SC_NOT_FOUND;

@RestController
@RequestScope
@RequestMapping("/")
@Validated
public class ActionApi {

    @Autowired
    private IActionRepo repo;
    @Autowired
    private TestActionService testActionService;
    @Autowired
    private AuditLog auditLog;
    @Autowired
    private JaxRsDpsLog log;
    @Autowired
    private IRetrieveActionService retrieveActionService;

    @PostMapping("action")
    @PreAuthorize("@authorizationFilter.hasAnyPermission('" + ServiceRole.OPS + "', '" + ServiceRole.ADMIN + "')")
    public ResponseEntity<Action> createAction(@RequestBody @Valid CreateActionDto createActionDto) {
        Action action = new Action(createActionDto);
        Action output = repo.createAction(action);
        auditLog.createActionSuccess(action);
        return new ResponseEntity<>(output, HttpStatus.CREATED);
    }

    @PostMapping("action:retrieve")
    @PreAuthorize("@authorizationFilter.hasAnyPermission('" + ServiceRole.OPS + "', '" + ServiceRole.ADMIN + "', '" + ServiceRole.EDITOR + "', '" + ServiceRole.VIEWER + "')")
    public ResponseEntity<List<Action>> retrieveAction(@RequestBody JsonNode jsonObject) {
        try {
            List<Action> query = repo.getAllActions();
            List<Action> output = this.retrieveActionService.getActions(query, jsonObject);
            return new ResponseEntity<>(output, HttpStatus.OK);
        } catch (Exception e) {
            this.log.error("retrieve action failed", e);
            return new ResponseEntity<>(HttpStatus.INTERNAL_SERVER_ERROR);
        }
    }

    @DeleteMapping("action/{id}")
    @PreAuthorize("@authorizationFilter.hasAnyPermission('" + ServiceRole.OPS + "', '" + ServiceRole.ADMIN + "')")
    public ResponseEntity deleteAction(@PathVariable("id") @NotBlank @Pattern(regexp = "^[A-Za-z0-9-]{2,50}") String id) {
        if (repo.delete(id)) {
            auditLog.deleteActionSuccess(id);
            return new ResponseEntity(HttpStatus.NO_CONTENT);
        } else {
            throw new AppException(SC_NOT_FOUND, "Not found", String.format("Could not delete Action with id %s because it could not be found", id));
        }
    }

    @GetMapping("action/{id}")
    @PreAuthorize("@authorizationFilter.hasAnyPermission('" + ServiceRole.OPS + "', '" + ServiceRole.ADMIN + "', '" + ServiceRole.EDITOR + "', '" + ServiceRole.VIEWER + "')")
    public ResponseEntity<Action> getAction(@PathVariable("id") @NotBlank @Pattern(regexp = "^[A-Za-z0-9-]{2,50}") String id) {
        return new ResponseEntity<>(repo.get(id), HttpStatus.OK);
    }

    @PostMapping("action:test")
    @PreAuthorize("@authorizationFilter.hasAnyPermission('" + ServiceRole.OPS + "', '" + ServiceRole.ADMIN + "', '" + ServiceRole.EDITOR + "', '" + ServiceRole.VIEWER + "')")
    public ResponseEntity<ParsedAction> testAction(@RequestBody @Valid TestActionRequest testActionRequest) {
        ParsedAction response = this.testActionService.validateRegex(testActionRequest);
        return new ResponseEntity<>(response, HttpStatus.OK);
    }
}
