/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.api;

import com.google.api.client.googleapis.auth.oauth2.GoogleIdToken;
import com.google.api.client.googleapis.auth.oauth2.GoogleIdTokenVerifier;
import com.google.api.client.http.javanet.NetHttpTransport;
import com.google.api.client.json.jackson2.JacksonFactory;
import com.google.common.base.Strings;
import com.google.common.hash.Hashing;
import org.opengroup.osdu.core.common.cryptographic.ISignatureService;
import org.opengroup.osdu.core.common.cryptographic.SignatureServiceException;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.register.utils.AppServiceConfig;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.validation.annotation.Validated;
import org.springframework.web.bind.annotation.*;
import org.springframework.web.context.annotation.RequestScope;

import javax.inject.Inject;
import javax.validation.constraints.NotBlank;
import java.nio.charset.StandardCharsets;
import java.util.Base64;
import java.util.Collections;

//used as both documentation and by integration test to validate challenge response logic
@RestController
@RequestMapping("/test")
@RequestScope
@Validated
public class SubscriberTestListenerApi {

    @Autowired
    private AppServiceConfig serviceConfig;
    @Autowired
    private ISignatureService signatureService;
    @Autowired
    private JaxRsDpsLog logger;
    @Inject
    private javax.inject.Provider<DpsHeaders> headersProvider;

    @GetMapping("challenge/{count}")
    public ResponseEntity<?> testCrc(@RequestParam("crc") @NotBlank String crc, @RequestParam("hmac") @NotBlank String hmac) {
        try {
            signatureService.verifyHmacSignature(hmac, this.serviceConfig.getSubscriberSecret());
        } catch (SignatureServiceException e) {
            return new ResponseEntity<>("Authorization signature validation Failed", HttpStatus.BAD_REQUEST);
        }
        logger.info("Signature verified and sending response");
        // Use the secret you send to the subscriber registration create request
        return getResponse(crc, this.serviceConfig.getSubscriberSecret());
    }

    @PostMapping("challenge/{count}")
    public ResponseEntity testPushHmac(@RequestBody Object o, @RequestParam("hmac") String hmac) {
        try {
            signatureService.verifyHmacSignature(hmac, this.serviceConfig.getSubscriberSecret());
        } catch (SignatureServiceException e) {
            return new ResponseEntity<>("Authorization signature validation Failed", HttpStatus.BAD_REQUEST);
        }
        logger.info("Sending acknowledgement from hmac endpoint");
        return new ResponseEntity<>(HttpStatus.OK);
    }

    @GetMapping("gsa-challenge/{count}")
    public ResponseEntity<?> testGsa(@RequestParam("crc") @NotBlank String crc) {
        if (!verifyToken()) {
            return new ResponseEntity<>("Authorization signature validation Failed", HttpStatus.BAD_REQUEST);
        }

        logger.info("Token verified and sending response");
        // google audience input is used as the secret for subscriber registration create request
        String secret = this.serviceConfig.getSubscriberPrivateKeyId();
        return getResponse(crc, secret);
    }


    @PostMapping("gsa-challenge/{count}")
    public ResponseEntity testPushGsa(@RequestBody Object o) {
        if (!verifyToken()) {
            return new ResponseEntity<>("Authorization signature validation Failed", HttpStatus.BAD_REQUEST);
        }

        logger.info("Sending acknowledgement from gsa endpoint");
        return new ResponseEntity<>(HttpStatus.OK);
    }

    class ChallengeResponse {
        public String responseHash = "";
    }

    private boolean verifyToken() {
        DpsHeaders headers = headersProvider.get();
        if (Strings.isNullOrEmpty(headers.getAuthorization()))
            return true;

        try {
            GoogleIdTokenVerifier verifier = new GoogleIdTokenVerifier.Builder(new NetHttpTransport(), JacksonFactory.getDefaultInstance())
                    .setAudience(Collections.singletonList(this.serviceConfig.getIntegrationTestJwtAudiences()))
                    .build();
            GoogleIdToken idToken = verifier.verify(headers.getAuthorization());
            return idToken != null;
        } catch (Exception ex) {
            return false;
        }
    }

    private ResponseEntity<ChallengeResponse> getResponse(String crc, String secretString) {
        String response = secretString + crc;
        response = Hashing.sha256()
                .hashString(response, StandardCharsets.UTF_8)
                .toString();
        response = Base64.getEncoder().encodeToString(response.getBytes());
        ChallengeResponse cr = new ChallengeResponse();
        cr.responseHash = response;
        return new ResponseEntity<>(cr, HttpStatus.OK);
    }
}
