/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.ddms.model.validators;

import com.google.common.base.Strings;
import com.reprezen.jsonoverlay.Overlay;
import com.reprezen.kaizen.oasparser.model3.OpenApi3;
import com.reprezen.kaizen.oasparser.val.ValidationContext;
import com.reprezen.kaizen.oasparser.val.ValidationResults;
import org.opengroup.osdu.register.ddms.model.ExtensionValidationMessages;
import org.opengroup.osdu.register.ddms.model.RegisteredInterface;
import org.opengroup.osdu.core.common.model.http.AppException;

import javax.validation.ConstraintValidator;
import javax.validation.ConstraintValidatorContext;

import static com.reprezen.kaizen.oasparser.val.msg.Messages.msg;

public class OpenApi3SpecValidator implements ConstraintValidator<ValidOpenApi3Spec, RegisteredInterface> {

    @Override
    public void initialize(ValidOpenApi3Spec constraintAnnotation) {
        //needed by interface - we don't use
    }

    @Override
    public boolean isValid(RegisteredInterface registeredInterface, ConstraintValidatorContext context) {
        try {
            ValidationResults results = validateOpenApiSpec(registeredInterface);
            OpenApi3 openApi3 = registeredInterface.openApi3Spec();
            openApi3.getValidationResults().add(results);
            for (ValidationResults.ValidationItem validationItem : openApi3.getValidationItems()) {
                context.disableDefaultConstraintViolation();
                context.buildConstraintViolationWithTemplate(validationItem.getMsg()).addConstraintViolation();
            }
            return openApi3.isValid();
        } catch (Exception e) {
            throw new AppException(500, "Server error", "Unexpected error parsing schema", e);
        }
    }

    private ValidationResults validateOpenApiSpec(RegisteredInterface registeredInterface) throws Exception {
        ValidationResults results;

        String server = registeredInterface.retrieveServerBaseUrl();
        OpenApi3 spec = registeredInterface.openApi3Spec();

        try (ValidationContext validationContext = ValidationContext.open()) {
            results = ValidationContext.getValidationResults();

            if (Strings.isNullOrEmpty(server)) {
                results.addError(msg(ExtensionValidationMessages.InvalidServers), Overlay.of(spec));
            }
        }
        return results;
    }
}
