/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.middleware;

import org.apache.commons.lang3.StringUtils;
import org.opengroup.osdu.core.common.model.entitlements.AuthorizationResponse;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.model.http.DpsHeaders;
import org.opengroup.osdu.core.common.provider.interfaces.IAuthorizationService;
import org.opengroup.osdu.register.utils.ServiceRequestInfo;
import org.opengroup.osdu.register.utils.ServiceRole;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.springframework.web.context.annotation.RequestScope;

import javax.servlet.http.HttpServletRequest;
import java.util.Arrays;

@Component("authorizationFilter")
@RequestScope
public class AuthorizationFilter {

    @Autowired
    private DpsHeaders headers;
    @Autowired
    private IAuthorizationService authorizationService;
    @Autowired
    private HttpServletRequest request;
    @Autowired
    private ServiceRequestInfo requestInfo;

    public boolean hasAnyPermission(String... requiredRoles) {
        String path = request.getServletPath();

        if ("GET".equals(request.getMethod())) {
            if (path.equals("/swagger-ui.html")) {
                return true;
            }
        }

        if (StringUtils.isBlank(this.headers.getAuthorization())) {
            throw new AppException(400, "Bad request", "No credentials sent on request.");
        }

        try {
            AuthorizationResponse response = checkApiAccess(requiredRoles, headers);
            headers.addCorrelationIdIfMissing();
            headers.put(DpsHeaders.USER_EMAIL, response.getUser());
        } catch (AppException e) {
            if (Arrays.asList(requiredRoles).contains(ServiceRole.CRON)) {
                checkCronAccess();
                headers.put(DpsHeaders.USER_EMAIL, ServiceRole.CRON); //just assign user as cron job
                return true;
            } else {
                return false;
            }
        }
        return true;
    }

    private void checkCronAccess() {
        if (!requestInfo.isCronRequest()) {
            throw new AppException(401, "Unauthorized", "The user is not authorized to perform this action");
        }
    }

    private AuthorizationResponse checkApiAccess(String[] requiredRoles, DpsHeaders dpsHeaders) {
        return authorizationService.authorizeAny(dpsHeaders, requiredRoles);
    }
}
