/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.subscriber.model.validators;

import com.google.common.base.Strings;
import org.opengroup.osdu.register.subscriber.model.GsaSecret;
import org.opengroup.osdu.register.subscriber.model.HmacSecret;
import org.opengroup.osdu.register.subscriber.model.Secret;
import org.opengroup.osdu.register.utils.Constants;

import javax.validation.ConstraintValidator;
import javax.validation.ConstraintValidatorContext;

public class SecretValidator implements ConstraintValidator<ValidSecret, Secret> {

    private static final String HMAC_VALUE_PATTERN = "^[a-zA-Z0-9]{8,30}+$";

    @Override
    public void initialize(ValidSecret constraintAnnotation) { }

    @Override
    public boolean isValid(Secret value, ConstraintValidatorContext context) {
        return validateSecretType(value.getSecretType(), context) && validateSecretValue(value, context);
    }

    private boolean validateSecretType(String secretType, ConstraintValidatorContext context) {
        if (secretType == null) {
            return addError(context, "Secret type not specified.");
        }

        if (secretType.equalsIgnoreCase(Constants.HMACSecret) || secretType.equalsIgnoreCase(Constants.GSASecret))
            return true;

        return addError(context, "Only 'HMAC' or 'GSA' types allowed");
    }

    private boolean validateSecretValue(Secret secret, ConstraintValidatorContext context) {
        if (secret.getSecretType().equalsIgnoreCase(Constants.HMACSecret)) {
            String hmacSecret = ((HmacSecret)secret).getValue();
            if(!hmacSecret.matches(HMAC_VALUE_PATTERN) || hmacSecret.length()% 2 != 0 )
                return addError(context, "HMAC value must match \"^[a-zA-Z0-9]{8,30}+$\" and length should be even.");
        } else {
            GsaSecret gsaSecret = (GsaSecret)secret;
            if(Strings.isNullOrEmpty(gsaSecret.getValue().getAudience()) || Strings.isNullOrEmpty(gsaSecret.getValue().getKey())) {
                return addError(context, "audience & key are required for GSA secret type");
            }
        }
        return true;
    }

    boolean addError(ConstraintValidatorContext context, String error){
        context.disableDefaultConstraintViolation();
        context.buildConstraintViolationWithTemplate(error).addConstraintViolation();
        return false;
    }
}
