/*
 * Copyright 2017-2020, Schlumberger
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.opengroup.osdu.register.action.services;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.opengroup.osdu.register.action.model.ParsedAction;
import org.opengroup.osdu.core.common.model.http.AppException;
import org.opengroup.osdu.core.common.logging.JaxRsDpsLog;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;

import java.util.HashMap;
import java.util.Map;

import static org.junit.Assert.*;
import static org.mockito.Mockito.verify;

@RunWith(MockitoJUnitRunner.class)
public class ParseJsonIntoActionTest {

    @InjectMocks
    private ParseJsonIntoAction sut;

    @Mock
    private JaxRsDpsLog log;

    @Test
    public void should_registerAction_given_URLHasNoTemplate() {
        String url = "https://myapp.osdu.opengroup.org/action/";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/", response.getUrl());
        assertEquals(expectedError, response.getErrors());
        verify(this.log).info("No template found on action.url: https://myapp.osdu.opengroup.org/action/");
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasEmptyTemplate() {
        String url = "https://myapp.osdu.opengroup.org/action/{}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Bad template format: empty template.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{}", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasInvalidTemplate_MissingSeparatingColon() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.uri^(?:.+\\\\:)*(.+)}/abc";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Bad template format: template variable and regex should be separated by colon.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{data.uri^(?:.+\\\\:)*(.+)}/abc", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasInvalidTemplate_MissingVariableName() {
        String url = "https://myapp.osdu.opengroup.org/action/{:^(?:.+\\\\:)*(.+)}/abc";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Bad template format: template should contain a variable name.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{:^(?:.+\\\\:)*(.+)}/abc", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasInvalidTemplate_MissingRegex() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.uri:}/abc";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Bad template format: provide regex or remove the colon.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{data.uri:}/abc", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasInvalidTemplate_MalformatTemplate() {
        String url = "https://myapp.osdu.opengroup.org/action/{id={id}/abc";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Bad template format:　no closing curly brace found for template.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{id={id}/abc", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasValidTemplate_NoRegex() {
        String url = "https://myapp.osdu.opengroup.org/action/{id}/abc";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/unit-test-id/abc", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasValidTemplate_RegexHasNoCurly() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.uri:[\\:\\s*(\\w+)]*}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/https:", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasValidTemplate_RegexHasPairsOfCurly() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.link:^(?:[^:]*:){1}([^:]*)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/type", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasValidTemplate_RegexHasExactCharacterCurly() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.withCurly:^(\\w+\\{)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/common{", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasValidTemplate_TemplateInQuery() {
        String url = "https://myapp.osdu.opengroup.org/action?projectId={data.link:^(?:.+\\:)*(.+)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action?projectId=100566", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasMultiPleTemplate_OneTemplateInvalid() {
        String url = "https://myapp.osdu.opengroup.org/action/{data:}?projectId={data.link:^(?:.+\\:)*(.+)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Bad template format: provide regex or remove the colon.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{data:}?projectId={data.link:^(?:.+\\:)*(.+)}", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasMultiPleTemplate_OneRegexInvalid() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.uri:^(??\\w+\\{)}?projectId={data.link:^(?:.+\\:)*(.+)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Invalid regex: ^(??\\w+\\{) found for template var: {data.uri:^(??\\w+\\{)}";
        assertEquals("https://myapp.osdu.opengroup.org/action/{data.uri:^(??\\w+\\{)}?projectId={data.link:^(?:.+\\:)*(.+)}", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_returnOriginalURLAndError_given_URLHasMultiPleTemplate_OneRegexHasNoMatch() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.uri:(\\})}?projectId={data.link:^(?:.+\\:)*(.+)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "Regex: (\\}) failed to extract a value from the test payload.";
        assertEquals("https://myapp.osdu.opengroup.org/action/{data.uri:(\\})}?projectId={data.link:^(?:.+\\:)*(.+)}", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasMultiPleTemplate_TwoTemplatesHasTheSameVariable() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.link:^(?:[^\\:]*\\:){1}([^\\:]*)}?projectId={data.link:^(?:.+\\:)*(.+)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/type?projectId=100566", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test
    public void should_registerAction_given_URLHasMultiPleTemplate_TwoTemplatesHasDifferentVariable() {
        String url = "https://myapp.osdu.opengroup.org/action/{kind:^(?:[^\\:]*\\:){1}([^\\:]*)}?projectId={data.link:^(?:.+\\:)*(.+)}";
        JsonNode testPayload = createTestPayload();

        ParsedAction response = this.sut.parse(url, testPayload);

        String expectedError = "";
        assertEquals("https://myapp.osdu.opengroup.org/action/petrel?projectId=100566", response.getUrl());
        assertEquals(expectedError, response.getErrors());
    }

    @Test(expected = AppException.class)
    public void should_throwAppException_given_RegexBackTracking() {
        String url = "https://myapp.osdu.opengroup.org/action/{data.timeout:(x+x+)+y}";
        JsonNode testPayload = createTestPayload();

        this.sut.parse(url, testPayload);
    }

    private JsonNode createTestPayload() {
        Map<String, Object> data = new HashMap<>();
        data.put("uri", "https://myproj.com/abc123");
        data.put("link", "common:type:unit-test-abc123:100566");
        data.put("withCurly", "common{unit");
        data.put("timeout", "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx");
        Map<String, Object> record = new HashMap<>();
        record.put("id", "unit-test-id");
        record.put("kind", "common:petrel:regularheightfield:1.0.0");
        record.put("data", data);
        ObjectMapper mapper = new ObjectMapper();
        return mapper.convertValue(record, JsonNode.class);
    }
}